# -*- coding: utf-8 -*-

import typing as T
import os
import json
import dataclasses


@dataclasses.dataclass
class BaseEnvVar:
    """
    A base class for environment secrets data container.
    """

    @classmethod
    def load(cls):
        kwargs = {f.name: os.environ.get(f.name) for f in dataclasses.fields(cls)}
        return cls(**kwargs)

    def to_json(self, pretty: bool=True) -> str:
        kwargs = dict()
        if pretty:
            kwargs["indent"] = 4
        return json.dumps(dataclasses.asdict(self), **kwargs)


@dataclasses.dataclass
class AWSEnvVar(BaseEnvVar):
    """
    AWS related environment variable secrets.

    Ref:

    - https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-envvars.html
    """

    AWS_ACCESS_KEY_ID: T.Optional[str] = dataclasses.field(default=None)
    AWS_CA_BUNDLE: T.Optional[str] = dataclasses.field(default=None)
    AWS_CLI_AUTO_PROMPT: T.Optional[str] = dataclasses.field(default=None)
    AWS_CLI_FILE_ENCODING: T.Optional[str] = dataclasses.field(default=None)
    AWS_CONFIG_FILE: T.Optional[str] = dataclasses.field(default=None)
    AWS_DATA_PATH: T.Optional[str] = dataclasses.field(default=None)
    AWS_DEFAULT_OUTPUT: T.Optional[str] = dataclasses.field(default=None)
    AWS_DEFAULT_REGION: T.Optional[str] = dataclasses.field(default=None)
    AWS_EC2_METADATA_DISABLED: T.Optional[str] = dataclasses.field(default=None)
    AWS_MAX_ATTEMPTS: T.Optional[str] = dataclasses.field(default=None)
    AWS_METADATA_SERVICE_NUM_ATTEMPTS: T.Optional[str] = dataclasses.field(default=None)
    AWS_METADATA_SERVICE_TIMEOUT: T.Optional[str] = dataclasses.field(default=None)
    AWS_PAGER: T.Optional[str] = dataclasses.field(default=None)
    AWS_PROFILE: T.Optional[str] = dataclasses.field(default=None)
    AWS_REGION: T.Optional[str] = dataclasses.field(default=None)
    AWS_RETRY_MODE: T.Optional[str] = dataclasses.field(default=None)
    AWS_ROLE_ARN: T.Optional[str] = dataclasses.field(default=None)
    AWS_ROLE_SESSION_NAME: T.Optional[str] = dataclasses.field(default=None)
    AWS_SECRET_ACCESS_KEY: T.Optional[str] = dataclasses.field(default=None)
    AWS_SESSION_TOKEN: T.Optional[str] = dataclasses.field(default=None)
    AWS_SHARED_CREDENTIALS_FILE: T.Optional[str] = dataclasses.field(default=None)
    AWS_USE_FIPS_ENDPOINT: T.Optional[str] = dataclasses.field(default=None)
    AWS_WEB_IDENTITY_TOKEN_FILE: T.Optional[str] = dataclasses.field(default=None)



# -*- coding: utf-8 -*-

import sys
from .covtest import run_cov_test
from .paths import dir_tests

py_ver = f"{sys.version_info.major}{sys.version_info.minor}"

try:
    from boto_session_manager import BotoSesManager

    bsm = BotoSesManager()
    bucket = f"{bsm.aws_account_id}-{bsm.aws_region}-data"
except ImportError as e: # pragma: no cover
    pass



# -*- coding: utf-8 -*-

"""
pysecret is a tiny library to allow developer load secret information safely.

- from environment variable
- from json file
- from AWS Secret Manager and Key Management Service
"""

from ._version import __version__

__short_description__ = "utility tool that load secret information safely."
__license__ = "MIT"
__author__ = "Sanhe Hu"
__author_email__ = "husanhe@gmail.com"
__github_username__ = "MacHu-GWU"

try:
    from .paths import (
        dir_home,
        path_bash_profile,
        path_bashrc,
        path_zshrc,
    )
    from .env import (
        BaseEnvVar,
        AWSEnvVar,
    )
    from .js import (
        JsonSecret,
        DEFAULT_JSON_SECRET_FILE,
    )
    from .sh import BaseShellScriptSecret
except ImportError:  # pragma: no cover
    pass
except Exception as e:  # pragma: no cover
    raise e

try:
    from .aws import (
        Parameter,
        deploy_parameter,
        delete_parameter,
        Secret,
        deploy_secret,
        delete_secret,
        kms_symmetric_encrypt,
        kms_symmetric_decrypt,
    )
except ImportError:  # pragma: no cover
    pass
except Exception as e:  # pragma: no cover
    raise e


def __getattr__(name: str):  # pragma: no cover
    if name in [
        "EnvSecret",
        "get_home_path",
        "AWSSecret",
    ]:
        raise AttributeError(
            f"The pysecret.{name} API has been removed since 2.X! "
            f"You can either downgrade to 1.0.4 or update your code."
        )
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")



# -*- coding: utf-8 -*-

import sys

if (
    sys.version_info.major == 2
    or (sys.version_info.major == 3 and sys.version_info.minor < 6)
):
    raise NotImplementedError("we don't support < Python3.6!")

if sys.version_info.minor < 8:
    from cached_property import cached_property
else:
    from functools import cached_property



# -*- coding: utf-8 -*-

from .parameter_store import (
    Parameter,
    deploy_parameter,
    delete_parameter,
)
from .secret_manager import (
    Secret,
    deploy_secret,
    delete_secret,
)
from .kms import (
    kms_symmetric_encrypt,
    kms_symmetric_decrypt,
)



# -*- coding: utf-8 -*-

import typing as T


def ensure_only_one_true(kv_list: T.List[T.Tuple[str, bool]]):
    if sum([v for _, v in kv_list]) != 1:
        raise ValueError(
            "one and only one argument in {} has to be True!".format(
                ", ".join([f"{k!r}" for k, _ in kv_list])
            )
        )



# -*- coding: utf-8 -*-

"""
This module provides high level API for AWS Key Management Service and
AWS Secret Manager.
"""

import base64
import json

import boto3

try:
    import typing
except:
    pass
from ..js_helper import get_value, strip_comments

JSON_PICKLE_KEY = "__jsonpickle__"


class AWSSecret(object):
    """
    AWS based secret. Support following backend:

    - AWS Key Management Service
    - AWS Parameter Store
    - AWS Secret Manager
    """

    def __init__(
        self,
        aws_access_key_id=None,
        aws_secret_access_key=None,
        aws_session_token=None,
        region_name=None,
        profile_name=None,
        boto_session=None,
    ):
        if boto_session is None:
            self.ses = boto3.Session(
                aws_access_key_id=aws_access_key_id,
                aws_secret_access_key=aws_secret_access_key,
                aws_session_token=aws_session_token,
                region_name=region_name,
                profile_name=profile_name,
            )
        else:
            self.ses = boto_session
        self._kms_client = None
        self._sm_client = None
        self._ssm_client = None
        self.secret_cache = dict()
        self.parameter_cache = dict()

    class KMSArgValueEnum:
        class KeyId:
            aws_ssm = "alias/aws/ssm"

    class UpdateModeEnum:
        create = 1
        upsert = 2
        try_create = 3

    class ParameterStoreArgValueEnum:
        class Tier:
            standard = "Standard"
            advanced = "Advanced"
            intelligent_tiering = "Intelligent-Tiering"

    @property
    def kms_client(self):  # Key Management Service client
        if self._kms_client is None:
            self._kms_client = self.ses.client("kms")
        return self._kms_client

    @property
    def sm_client(self):  # Secret Manager client
        if self._sm_client is None:
            self._sm_client = self.ses.client("secretsmanager")
        return self._sm_client

    @property
    def ssm_client(self):  # System Manager client
        if self._ssm_client is None:
            self._ssm_client = self.ses.client("ssm")
        return self._ssm_client

    # --------------------------------------------------------------------------
    # AWS Parameter Store integration
    # --------------------------------------------------------------------------
    def _put_parameter(self, parameter_name, put_parameter_kwargs):
        """
        :type parameter_name: str
        :type put_parameter_kwargs: dict

        :rtype: dict
        """
        response = self.ssm_client.put_parameter(**put_parameter_kwargs)
        self.parameter_cache[parameter_name] = None  # reset cache if upsert success
        return response

    def deploy_parameter(
        self,
        name,
        parameter_data,
        description=None,
        kms_key_id=None,
        use_default_kms_key=False,
        tier=None,
        policies=None,
        tags=None,
        update_mode=UpdateModeEnum.create,
        skip_duplicate=True,
    ):
        """
        Create or Update a parameter.

        - boto3 put_parameter api: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ssm.html#SSM.Client.put_parameter

        :type name: str
        :type parameter_data: dict
        :type description: str
        :type kms_key_id: str
        :type use_default_kms_key: bool
        :param use_default_kms_key: if true, use the default
            ``alias/aws/ssm`` kms key for encryption
        :type tier: str
        :type policies: str
        :type tags: typing.Dict[str, str]
        :type update_mode: int
        :type skip_duplicate: bool

        :rtype: dict

        Update mode:

        1. create: create new parameter, raise error if already exists
        2. upsert: create new parameter, overwrite existing one if already exists,
            if skip_duplicate is True, then won't do deployment if parameter data
            not changed
        3. try_create: create new parameter, silently do nothing if already exists

        update mode code can be accessed from ``AWSSecret.UpdateModeEnum``
        """
        if tags is None:
            tags = dict()
        Tags = [{"Key": k, "Value": v} for k, v in tags.items()]

        put_parameter_kwargs = {
            "Name": name,
            "Value": json.dumps(parameter_data),
            "Type": "String",
            "Tier": "Intelligent-Tiering",
        }
        if description is not None:
            put_parameter_kwargs["Description"] = description
        if (use_default_kms_key is True) and (kms_key_id is not None):
            msg = (
                "you cannot set `use_default_kms_key` and " "`kms_key_id` at same time!"
            )
            raise ValueError(msg)
        if use_default_kms_key:
            kms_key_id = self.KMSArgValueEnum.KeyId.aws_ssm
        if kms_key_id is not None:
            put_parameter_kwargs["Type"] = "SecureString"
            put_parameter_kwargs["KeyId"] = kms_key_id
        if policies is not None:
            put_parameter_kwargs["Policies"] = policies
        if tier is not None:
            put_parameter_kwargs["Tier"] = tier
        if len(Tags):
            put_parameter_kwargs["Tags"] = Tags

        if update_mode == self.UpdateModeEnum.create:
            put_parameter_kwargs["Overwrite"] = False
            return self._put_parameter(name, put_parameter_kwargs)
        elif update_mode == self.UpdateModeEnum.upsert:
            put_parameter_kwargs["Overwrite"] = True
            if skip_duplicate:
                try:
                    if (kms_key_id is not None) or (use_default_kms_key is True):
                        with_encryption = True
                    else:
                        with_encryption = False
                    response = self.ssm_client.get_parameter(
                        Name=name,
                        WithDecryption=with_encryption,
                    )
                    if "Parameter" in response:
                        string_value = response["Parameter"]["Value"]
                    else:  # pragma: no cover
                        raise ValueError("Not a valid get_parameter response!")
                    data = json.loads(strip_comments(string_value))
                    if parameter_data == data:  # duplicate parameter data
                        return {}
                    else:
                        return self._put_parameter(name, put_parameter_kwargs)
                except Exception as e:
                    if "ParameterNotFound" in str(e):
                        return self._put_parameter(name, put_parameter_kwargs)
                    else:
                        raise e
            else:
                return self._put_parameter(name, put_parameter_kwargs)
        elif update_mode == self.UpdateModeEnum.try_create:
            try:
                return self._put_parameter(name, put_parameter_kwargs)
            except Exception as e:
                if "exists" in str(e).lower():
                    return {}
                else:
                    raise e
        else:
            raise ValueError("{} is not a valid update_mode code".format(update_mode))

    def get_parameter_raw_value(
        self,
        name,
        with_encryption=False,
    ):
        """
        Get the raw parameter data in form of string.

        :type name: str
        :param name: aws system manager parameter name

        :type with_encryption: bool
        :param with_encryption: use True if it is a secure string type parameter

        :rtype: str
        """
        response = self.ssm_client.get_parameter(
            Name=name,
            WithDecryption=with_encryption,
        )
        if "Parameter" in response:
            return response["Parameter"]["Value"]
        else:  # pragma: no cover
            raise ValueError("Not a valid get_parameter response!")

    def get_parameter_data(
        self,
        name,
        with_encryption=False,
    ):
        """
        Get the parameter data in form of json dictionary.

        :type name: str
        :param name: aws system manager parameter name

        :type with_encryption: bool
        :param with_encryption: use True if it is a secure string type parameter

        :rtype: dict
        """
        if self.parameter_cache.get(name) is None:
            response = self.ssm_client.get_parameter(
                Name=name,
                WithDecryption=with_encryption,
            )
            if "Parameter" in response:
                string_value = response["Parameter"]["Value"]
            else:  # pragma: no cover
                raise ValueError("Not a valid get_parameter response!")
            data = json.loads(strip_comments(string_value))
            self.parameter_cache[name] = data
        else:
            data = self.parameter_cache[name]
        return data

    def get_parameter_value(
        self,
        name,
        json_path,
        with_encryption=False,
    ):
        """
        Fetch a specific value using json path dot notation

        :type name: str
        :param name: aws system manager parameter name

        :type json_path: str
        :param json_path: json path dot notation

        :type with_encryption: bool
        :param with_encryption: use True if it is a secure string type parameter

        :rtype: str
        :return: parameter value in string
        """
        data = self.get_parameter_data(name, with_encryption)
        return get_value(data, json_path)

    def deploy_parameter_object(
        self,
        name,
        parameter_obj,
        description=None,
        kms_key_id=None,
        use_default_kms_key=False,
        tier=None,
        policies=None,
        tags=None,
        update_mode=UpdateModeEnum.create,
        skip_duplicate=True,
    ):
        """
        Deploy a parameter object implemented using attrs library https://pypi.org/project/attrs/
        to AWS parameter store. It use jsonpickle https://pypi.org/project/attrs/
        library for serialization.

        :type name: str
        :type parameter_obj: typing.Any
        :type description: str
        :type kms_key_id: str
        :type use_default_kms_key: bool
        :param use_default_kms_key: if true, use the default
            ``alias/aws/ssm`` kms key for encryption
        :type tier: str
        :type policies: str
        :type tags: typing.Dict[str, str]
        :type update_mode: int
        :type skip_duplicate: bool

        :rtype: dict

        Update mode:

        1. create: create new parameter, raise error if already exists
        2. upsert: create new parameter, overwrite existing one if already exists,
            if skip_duplicate is True, then won't do deployment if parameter data
            not changed
        3. try_create: create new parameter, silently do nothing if already exists
        """
        import jsonpickle

        parameter_data = {JSON_PICKLE_KEY: jsonpickle.dumps(parameter_obj)}
        return self.deploy_parameter(
            name=name,
            parameter_data=parameter_data,
            description=description,
            kms_key_id=kms_key_id,
            use_default_kms_key=use_default_kms_key,
            tier=tier,
            policies=policies,
            tags=tags,
            update_mode=update_mode,
            skip_duplicate=skip_duplicate,
        )

    def get_parameter_object(
        self,
        name,
        with_encryption=False,
    ):
        import jsonpickle

        data = self.get_parameter_data(
            name=name,
            with_encryption=with_encryption,
        )
        return jsonpickle.loads(data[JSON_PICKLE_KEY])

    def delete_parameter(
        self,
        name: str,
    ) -> dict:  # pragma: no cover
        return self.ssm_client.delete_parameter(Name=name)

    # --------------------------------------------------------------------------
    # AWS Key Management Service integration
    # --------------------------------------------------------------------------
    def kms_symmetric_encrypt(
        self,
        blob,
        kms_key_id,
    ):
        """
        Use KMS key to encrypt a short text.

        - KMS.Client.encrypt: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/kms.html#KMS.Client.encrypt

        :type text_blob: bytes
        :param text_blob:

        :type kms_key_id: str
        :param kms_key_id:

        :rtype: bytes
        """
        return self.kms_client.encrypt(
            Plaintext=blob,
            KeyId=kms_key_id,
        )["CiphertextBlob"]

    def kms_symmetric_decrypt(self, blob):
        """
        Use KMS key to decrypt a short text.

        :type blob: bytes
        :param blobtext: binary data to decrypt

        :rtype: bytes
        """
        return self.kms_client.decrypt(CiphertextBlob=blob)["Plaintext"]

    # --------------------------------------------------------------------------
    # AWS Secret Manager integration
    # --------------------------------------------------------------------------
    def deploy_secret(
        self,
        name,
        secret_data,
        description=None,
        kms_key_id=None,
        tags=None,
        update_mode=UpdateModeEnum.create,
    ):
        """
        Create or Update a AWS Secret.

        - create_secret: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.create_secret
        - update_secret: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.update_secret

        :type name: str
        :type secret_data: dict
        :type description: str
        :type kms_key_id: str
        :type tags: typing.Dict[str, str]
        :type update_mode: int

        Update mode:

        1. create: create new secret, raise error if already exists
        2. upsert: create new secret, overwrite existing one if exists
        3. try create: create new secret, silently do nothing if already exists

        Note:

            don't support Secret binary yet.
        """
        if tags is None:
            tags = dict()
        Tags = [{"Key": k, "Value": v} for k, v in tags.items()]

        create_or_update_secret_kwargs = {"SecretString": json.dumps(secret_data)}
        if description:
            create_or_update_secret_kwargs["Description"] = description
        if kms_key_id:
            create_or_update_secret_kwargs["KmsKeyId"] = kms_key_id
        if Tags:
            create_or_update_secret_kwargs["Tags"] = Tags

        try:  # create secret
            create_or_update_secret_kwargs["Name"] = name
            response = self.sm_client.create_secret(**create_or_update_secret_kwargs)
            self.secret_cache[name] = None
            return response
        except Exception as e:  # update secret
            if "exists" in str(e).lower():
                if update_mode == self.UpdateModeEnum.create:
                    raise e
                elif update_mode == self.UpdateModeEnum.upsert:
                    create_or_update_secret_kwargs.pop("Name")
                    create_or_update_secret_kwargs["SecretId"] = name
                    create_or_update_secret_kwargs.pop("Tags", None)
                    if len(tags):
                        print(
                            "ignore tags because you cannot update tag in `update_secret` API"
                        )
                    response = self.sm_client.update_secret(
                        **create_or_update_secret_kwargs
                    )
                    self.secret_cache[name] = None
                    return response
                elif update_mode == self.UpdateModeEnum.try_create:
                    return {}
                else:
                    raise ValueError(
                        "{} is not a valid update_mode code".format(update_mode)
                    )
            else:
                raise e

    def get_secret_raw_value(self, secret_id):
        """
        Get the raw secret data in form of string or binary.

        :type secret_id: str

        :rtype: typing.Union[str, bytes]
        """
        response = self.sm_client.get_secret_value(SecretId=secret_id)
        if "SecretString" in response:
            return response["SecretString"]
        else:
            return base64.b64decode(response["SecretBinary"])

    def get_secret_data(self, secret_id):
        """
        Get the secret data in form of json dictionary.

        :type secret_id: str
        :param secret_id: aws secret id

        :type key: str
        :param key: secret value dictionary key

        :rtype: dict
        """
        if self.secret_cache.get(secret_id) is None:
            response = self.sm_client.get_secret_value(SecretId=secret_id)
            if "SecretString" in response:
                secret = response["SecretString"]
            else:
                decoded_binary_secret = base64.b64decode(response["SecretBinary"])
                secret = decoded_binary_secret.decode("utf-8")
            data = json.loads(strip_comments(secret))
            self.secret_cache[secret_id] = data
        else:
            data = self.secret_cache[secret_id]
        return data

    def get_secret_value(self, secret_id, json_path):
        """
        Fetch a specific value using json path dot notation

        :type secret_id: str
        :param secret_id: aws secret id

        :type json_path: str
        :param json_path: json path dot notation
        """
        data = self.get_secret_data(secret_id)
        return get_value(data, json_path)

    def deploy_secret_object(
        self,
        name,
        secret_obj,
        description=None,
        kms_key_id=None,
        tags=None,
        update_mode=UpdateModeEnum.create,
    ):
        """
        Deploy a secret object implemented using attrs library https://pypi.org/project/attrs/
        to AWS Secret Manager. It use jsonpickle https://pypi.org/project/attrs/
        library for serialization.

        :type name: str
        :type secret_obj: typing.Any
        :type description: str
        :type kms_key_id: str
        :type tags: typing.Dict[str, str]
        :type update_mode: int

        :rtype: dict

        Update mode:

        1. create: create new parameter, raise error if already exists
        2. upsert: create new parameter, overwrite existing one if exists
        3. try create: create new parameter, silently do nothing if already exists
        """
        import jsonpickle

        secret_data = {JSON_PICKLE_KEY: jsonpickle.dumps(secret_obj)}
        return self.deploy_secret(
            name=name,
            secret_data=secret_data,
            description=description,
            kms_key_id=kms_key_id,
            tags=tags,
            update_mode=update_mode,
        )

    def get_secret_object(self, secret_id):
        import jsonpickle

        data = self.get_secret_data(secret_id)
        return jsonpickle.loads(data[JSON_PICKLE_KEY])

    def delete_secret(
        self,
        secret_id: str,
        recovery_window_in_days=30,
        force_delete_without_recovery=False,
    ) -> dict:  # pragma: no cover
        return self.sm_client.delete_secret(
            SecretId=secret_id,
            RecoveryWindowInDays=recovery_window_in_days,
            force_delete_without_recovery=force_delete_without_recovery,
        )



# -*- coding: utf-8 -*-

"""
AWS Secret Manager support.
"""

import typing as T
import json

import dataclasses
from datetime import datetime

from ..compat import cached_property
from ..js_helper import strip_comments


@dataclasses.dataclass
class Secret:
    """
    AWS Secret Manager secret object.

    - The camel case attributes are raw value from AWS API.
    - The snake case attributes are user-friendly accessor to the data.
    - only one of ``SecretBinary`` or ``SecretString`` could exist.
    - if you know what data type to expect in the secret, please use
        :meth:`Secret.binary`, :meth:`Secret.string`, :meth:`Secret.json_dict`,
        :meth:`Secret.json_list` to access the data.
    """
    ARN: str = dataclasses.field()
    Name: str = dataclasses.field()
    VersionId: str = dataclasses.field()
    CreatedDate: datetime = dataclasses.field()
    SecretBinary: T.Optional[bytes] = dataclasses.field(default=None)
    SecretString: T.Optional[str] = dataclasses.field(default=None)
    VersionStages: T.List[str] = dataclasses.field(default_factory=list)

    @cached_property
    def fingerprint(self) -> bytes:
        """
        The fingerprint of the content. Can be used for comparison.
        """
        if self.SecretBinary is not None:
            return self.SecretBinary
        else:
            return self.SecretString.encode("utf-8")

    @classmethod
    def load(
        cls,
        sm_client,
        name_or_arn: str,
        version_id: T.Optional[str] = None,
        version_stage: T.Optional[str] = None,
    ) -> T.Optional["Secret"]:
        """
        Load secret data.

        Ref:

        - describe_secret: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.describe_secret
        - get_secret_value: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.get_secret_value
        - list_secret_version_ids: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.list_secret_version_ids
        """
        # --- resolve arguments
        kwargs = dict(SecretId=name_or_arn)
        if version_id:  # pragma: no cover
            kwargs["VersionId"] = version_id
        if version_id:  # pragma: no cover
            kwargs["VersionStage"] = version_stage

        try:
            response = sm_client.get_secret_value(**kwargs)
            return cls(
                ARN=response["ARN"],
                Name=response["Name"],
                VersionId=response["VersionId"],
                SecretBinary=response.get("SecretBinary"),
                SecretString=response.get("SecretString"),
                CreatedDate=response["CreatedDate"],
                VersionStages=response.get("VersionStages", []),
            )
        except Exception as e:
            if "ResourceNotFoundException" in str(e):
                return None
            else:  # pragma: no cover
                raise e

    @classmethod
    def _from_create_or_update_secret_response(
        cls,
        create_or_update_secret_kwargs: dict,
        create_or_update_secret_response: dict,
    ):
        return Secret(
            ARN=create_or_update_secret_response["ARN"],
            Name=create_or_update_secret_response["Name"],
            VersionId=create_or_update_secret_response["VersionId"],
            SecretBinary=create_or_update_secret_kwargs.get("SecretBinary"),
            SecretString=create_or_update_secret_kwargs.get("SecretString"),
            CreatedDate=datetime.strptime(
                create_or_update_secret_response["ResponseMetadata"]["HTTPHeaders"][
                    "date"
                ],
                "%a, %d %b %Y %H:%M:%S %Z",
            ),
        )

    @property
    def binary(self) -> bytes:
        """
        The binary user data.
        """
        return self.SecretBinary

    @property
    def string(self) -> str:
        """
        The string user data.
        """
        return self.SecretString

    @cached_property
    def json_dict(self) -> dict:
        """
        The python dict user data.
        """
        return json.loads(strip_comments(self.SecretString))

    @cached_property
    def json_list(self) -> list:  # pragma: no cover
        """
        The python list user data.
        """
        return json.loads(strip_comments(self.SecretString))

    @property
    def aws_account_id(self) -> str:
        """
        The aws account id of this secret.
        """
        return self.ARN.split(":")[4]

    @property
    def aws_region(self) -> str:
        """
        The aws region of this secret.
        """
        return self.ARN.split(":")[3]


def deploy_secret(
    sm_client,
    name_or_arn: str,
    data: T.Union[bytes, str, list, dict, T.Any],
    description: T.Optional[str] = None,
    kms_key_id: T.Optional[str] = None,
    tags: T.Optional[T.Dict[str, str]] = None,
    add_replica_regions: T.Optional[T.List[T.Dict[str, str]]] = None,
    force_overwrite_replica_secret: T.Optional[bool] = None,
    client_request_token: T.Optional[str] = None,
    skip_if_duplicated: bool = True,
) -> T.Optional[Secret]:
    """
    Create or Update an AWS Secret.

    - create_secret: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.create_secret
    - update_secret: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.update_secret
    - tag_resource: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.tag_resource
    - untag_resource: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.untag_resource

    Note:

        secret manager can only add tag in creation, update_secret doesn't
        support tagging, this function will automatically call ``tag_resource``
        API when needed.

    :param sm_client: the boto3 secretmanager client.
    :param name_or_arn: name or the ARN of this secret.
    :param data: secret data you want to store, currently it supports bytes,
        string, json serializable dict or list.
    :param description: description of this secret.
    :param kms_key_id: the KMS key id you want to use for encryption, by default
        it uses the AWS managed KMS key.
    :param tags: the key value pair of the AWS resource tags.
    :param add_replica_regions: see official document.
    :param force_overwrite_replica_secret: see official document.
    :param client_request_token: see official document.
    :param skip_if_duplicated: default True, if True, will compare the secret data
        to the existing one before deployment. If they are the same, then
        no deployment happens.

    :return: None or an :class:`Secret` object, None means that the deployment
        doesn't happen.
    """
    # --------------------------------------------------------------------------
    # input argument pre processing
    # --------------------------------------------------------------------------
    create_or_update_secret_kwargs = dict()

    # data
    if isinstance(data, bytes):
        create_or_update_secret_kwargs["SecretBinary"] = data
    elif isinstance(data, str):
        create_or_update_secret_kwargs["SecretString"] = data
    elif isinstance(data, (list, dict)):
        create_or_update_secret_kwargs["SecretString"] = json.dumps(data)
    else:  # pragma: no cover
        raise NotImplementedError

    # description
    if description:
        create_or_update_secret_kwargs["Description"] = description

    # kms key id
    if kms_key_id:  # pragma: no cover
        create_or_update_secret_kwargs["KmsKeyId"] = kms_key_id

    # tags
    if tags is None:
        tags = dict()
    tags_ = [{"Key": k, "Value": v} for k, v in tags.items()]

    # --------------------------------------------------------------------------
    # create or update
    # --------------------------------------------------------------------------
    secret = Secret.load(
        sm_client,
        name_or_arn=name_or_arn,
    )
    is_create = secret is None

    # create branch
    if is_create:
        create_or_update_secret_kwargs["Name"] = name_or_arn
        if len(tags_):
            create_or_update_secret_kwargs["Tags"] = tags_
        if add_replica_regions is not None:  # pragma: no cover
            create_or_update_secret_kwargs["AddReplicaRegions"] = add_replica_regions
        if force_overwrite_replica_secret is not None:  # pragma: no cover
            create_or_update_secret_kwargs[
                "ForceOverwriteReplicaSecret"
            ] = add_replica_regions
        if client_request_token is not None:  # pragma: no cover
            create_or_update_secret_kwargs["ClientRequestToken"] = client_request_token
        response = sm_client.create_secret(**create_or_update_secret_kwargs)
        secret = Secret._from_create_or_update_secret_response(
            create_or_update_secret_kwargs=create_or_update_secret_kwargs,
            create_or_update_secret_response=response,
        )
        return secret

    # update branch
    # check duplication
    if skip_if_duplicated:
        # find existing secret's fingerprint
        if "SecretBinary" in create_or_update_secret_kwargs:
            fingerprint = create_or_update_secret_kwargs["SecretBinary"]
        else:
            fingerprint = create_or_update_secret_kwargs["SecretString"].encode("utf-8")
        # if the same, do nothing
        if fingerprint == secret.fingerprint:
            return None

    create_or_update_secret_kwargs["SecretId"] = name_or_arn
    response = sm_client.update_secret(**create_or_update_secret_kwargs)
    secret = Secret._from_create_or_update_secret_response(
        create_or_update_secret_kwargs=create_or_update_secret_kwargs,
        create_or_update_secret_response=response,
    )

    # do tagging
    if tags_ is not None:
        sm_client.tag_resource(SecretId=name_or_arn, Tags=tags_)

    return secret


def delete_secret(
    sm_client,
    name_or_arn: str,
    recovery_window_in_days: T.Optional[int] = None,
    force_delete_without_recovery: T.Optional[bool] = None,
) -> bool:
    """
    Delete a Secret.

    Ref:

    - delete_secret: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/secretsmanager.html#SecretsManager.Client.delete_secret

    :param sm_client: the boto3 secretmanager client.
    :param name_or_arn: name or the ARN of this secret.
    :param recovery_window_in_days: see official document.
    :param force_delete_without_recovery: see official document.

    :return: a boolean value to indicate whether a deletion happened.
    """
    kwargs = dict(SecretId=name_or_arn)
    if recovery_window_in_days is not None:  # pragma: no cover
        kwargs["RecoveryWindowInDays"] = recovery_window_in_days
    if force_delete_without_recovery is not None:
        kwargs["ForceDeleteWithoutRecovery"] = force_delete_without_recovery
    try:
        sm_client.delete_secret(**kwargs)
        return True
    except Exception as e:
        if "ResourceNotFoundException" in str(e):
            return False
        else:  # pragma: no cover
            raise e



# -*- coding: utf-8 -*-

import typing as T
import json
from pathlib import Path

from .js_helper import (
    create_json_if_not_exists,
    set_value,
    get_value,
    del_key,
    strip_comments,
)
from .singleton import CachedSpam

DEFAULT_JSON_SECRET_FILE = Path.home().joinpath(".pysecret.json")


class JsonSecret(CachedSpam):
    """
    Read and Write secret information from a JSON file.

    The secret content has to be a valid dictionary in JSON.
    """

    settings_uuid_field = "secret_file"

    def __real_init__(self, secret_file: Path = DEFAULT_JSON_SECRET_FILE):
        self.secret_file: Path = secret_file
        create_json_if_not_exists(str(self.secret_file))
        with open(self.secret_file, "rb") as f:
            self.data = json.loads(strip_comments(f.read().decode("utf-8")))

    def set(self, json_path: str, value) -> dict:
        self.data = set_value(self.data, json_path, value)
        self.secret_file.write_text(
            json.dumps(
                self.data,
                indent=4,
                ensure_ascii=False,
            )
        )
        return self.data

    def get(self, json_path: str) -> T.Any:
        return get_value(self.data, json_path)

    def unset(self, json_path: str):
        return del_key(self.data, json_path)



# -*- coding: utf-8 -*-

import typing as T
from pathlib import Path
from re import findall


def create_json_if_not_exists(path: str):
    """
    Create an empty json file if not exists
    """
    p = Path(path)
    if p.exists() is False:
        p.parent.mkdir(parents=True, exist_ok=True)
        p.write_text("{}")


def set_value(
    data: dict,
    json_path: str,
    value: T.Any,
) -> dict:
    """
    Set a field in dictionary data using JSON path syntax.

    :param data: the dictionary data
    :param json_path: the dot notation JSON path syntax, for example:
        ".key", ".key1.key2.key3"
    :param value: the changed data

    TODO: add array index support
    """
    if json_path.startswith("."):
        if json_path == ".":
            return value
        else:
            json_path = json_path[1:]

    paths = json_path.split(".")
    path_length = len(paths)
    dct = data
    for ind, key in enumerate(paths):
        if ind == (path_length - 1):
            dct[key] = value
        else:
            dct.setdefault(key, {})
            dct = dct[key]
    return data


def get_value(
    data: dict,
    json_path: str,
) -> T.Any:
    """
    Get the value of a field in dictionary data using JSON path syntax.

    :param data: the dictionary data
    :param json_path: the dot notation JSON path syntax, for example:

    TODO: add array index support
    """
    if json_path.startswith("."):
        if json_path == ".":
            return data
        else:
            json_path = json_path[1:]

    paths = json_path.split(".")
    value = data
    for key in paths:
        value = value[key]
    return value


def del_key(
    data: dict,
    json_path: str,
):
    """
    Delete a field in dictionary data using JSON path syntax.

    :param data: the dictionary data
    :param json_path: the dot notation JSON path syntax, for example:

    TODO: add array index support
    """
    paths = json_path.split(".")
    for key in paths[:-1]:
        data = data[key]
    del data[paths[-1]]


def strip_comment_line_with_symbol(
    line: str,
    comment_symbol: str,
) -> str:  # pragma: no cover
    """
    Strip comments from line string.

    :param line: the single line string that you want to strip out comments.
    :param comment_symbol: the comment char that indicate that the comment starts from.

    :return: the single line string with comment removed.
    """
    parts = line.split(comment_symbol)
    counts = [len(findall(r'(?:^|[^"\\]|(?:\\\\|\\")+)(")', part)) for part in parts]
    total = 0
    for nr, count in enumerate(counts):
        total += count
        if total % 2 == 0:
            return comment_symbol.join(parts[: nr + 1]).rstrip()
    else:  # pragma: no cover
        return line.rstrip()


def strip_comments(
    text: str,
    comment_symbols=frozenset(("#", "//")),
) -> str:  # pragma: no cover
    """
    Strip comments from json string.

    :param text: A string containing json with comments started by comment_symbols.
    :param comment_symbols: Iterable of symbols that start a line comment (default # or //).

    :return: the multi line text with the comments removed.
    """
    lines = text.splitlines()
    for k in range(len(lines)):
        for symbol in comment_symbols:
            lines[k] = strip_comment_line_with_symbol(lines[k], symbol)
    return "\n".join(lines)



# -*- coding: utf-8 -*-

import sys
from pathlib import Path

dir_project_root = Path(__file__).absolute().parent.parent.parent

# code structure
dir_tests = dir_project_root / "tests"
dir_htmlcov = dir_project_root / "htmlcov"
path_cov_index_html = dir_htmlcov / "index.html"

# virtual environment
dir_bin = Path(sys.executable).parent
bin_pytest = dir_bin / "pytest"



# -*- coding: utf-8 -*-

"""
AWS Parameter Store support.
"""

import typing as T
import json
import enum
import dataclasses
from datetime import datetime

try:
    import jsonpickle

    has_jsonpickle = True
except ImportError:  # pragma: no cover
    has_jsonpickle = False

from ..compat import cached_property
from ..js_helper import strip_comments
from ..helper import ensure_only_one_true

JSON_PICKLE_KEY = "__jsonpickle__"


class ParameterTypeEnum(str, enum.Enum):
    string = "String"
    string_list = "StringList"
    secure_string = "SecureString"


DEFAULT_KMS_KEY = "alias/aws/ssm"


class ParameterTierEnum(str, enum.Enum):
    standard = "Standard"
    advanced = "Advanced"
    intelligent_tiering = "Intelligent-Tiering"


class ParameterDataTypeEnum(str, enum.Enum):
    text = "text"
    ec2_image = "aws:ec2:image"


@dataclasses.dataclass
class Parameter:
    """
    AWS System Manager Parameter object.

    - The camel case attributes are raw value from AWS API.
    - The snake case attributes are user-friendly accessor to the data.
    - if you know what data type to expect in the parameter, please use
        :meth:`Parameter.string`, :meth:`Parameter.string_list`,
        :meth:`Parameter.json_dict`, :meth:`Parameter.json_list`,
        :meth:`Parameter.py_object` to access the data.
    """
    Name: str = dataclasses.field()
    Type: str = dataclasses.field()
    Value: str = dataclasses.field()
    Version: int = dataclasses.field()
    LastModifiedDate: datetime = dataclasses.field()
    DataType: str = dataclasses.field()
    ARN: T.Optional[str] = dataclasses.field(default=None)
    Selector: T.Optional[str] = dataclasses.field(default=None)
    SourceResult: T.Optional[str] = dataclasses.field(default=None)

    @classmethod
    def load(
        cls,
        ssm_client,
        name: str,
        with_decryption: T.Optional[bool] = None,
    ) -> T.Optional["Parameter"]:
        """
        Load parameter data.

        Ref:

        - get_parameter: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ssm.html#SSM.Client.get_parameter
        """
        kwargs = dict(Name=name)
        if with_decryption is not None:
            kwargs["WithDecryption"] = with_decryption

        # get the parameter data
        try:
            response = ssm_client.get_parameter(**kwargs)
            parameter = cls(
                Name=response["Parameter"]["Name"],
                Type=response["Parameter"]["Type"],
                Value=response["Parameter"]["Value"],
                Version=response["Parameter"]["Version"],
                LastModifiedDate=response["Parameter"]["LastModifiedDate"],
                DataType=response["Parameter"]["DataType"],
                ARN=response["Parameter"]["ARN"],
                Selector=response["Parameter"].get("Selector"),
                SourceResult=response["Parameter"].get("SourceResult"),
            )
            # check if the Type is secure string
            if parameter.Type == ParameterTypeEnum.secure_string.value:
                # if already set with_decryption = True, then return
                if with_decryption is True:
                    return parameter
                # if forget to set with_description = True, then do it again
                else:
                    return cls.load(ssm_client, name, True)
            else:
                return parameter
        # if not exists, return None
        except Exception as e:
            if "ParameterNotFound" in str(e):
                return None
            else:  # pragma: no cover
                raise e

    @classmethod
    def _from_put_parameter_response(
        cls,
        put_parameter_kwargs: dict,
        put_parameter_response: dict,
    ):
        return Parameter(
            Name=put_parameter_kwargs["Name"],
            Type=put_parameter_kwargs["Type"],
            Value=put_parameter_kwargs["Value"],
            Version=put_parameter_response["Version"],
            LastModifiedDate=datetime.strptime(
                put_parameter_response["ResponseMetadata"]["HTTPHeaders"]["date"],
                "%a, %d %b %Y %H:%M:%S %Z",
            ),
            ARN=None,
            DataType="text",
        )

    @property
    def string(self) -> str:
        """
        The string user data.
        """
        return self.Value

    @cached_property
    def string_list(self) -> T.List[str]:
        """
        The python string list user data.
        """
        return self.Value.split(",")

    @cached_property
    def json_dict(self) -> dict:
        """
        The python dict user data.
        """
        return json.loads(strip_comments(self.Value))

    @cached_property
    def json_list(self) -> list:
        """
        The python list user data.
        """
        return json.loads(strip_comments(self.Value))

    @cached_property
    def py_object(self):
        if has_jsonpickle is False:  # pragma: no cover
            raise ImportError(
                "you have to install `jsonpickle` to store arbitrary "
                "python object in AWS Parameter Store."
            )
        return jsonpickle.loads(json.loads(self.Value)[JSON_PICKLE_KEY])

    @property
    def aws_account_id(self) -> str:
        return self.ARN.split(":")[4]

    @property
    def aws_region(self) -> str:
        return self.ARN.split(":")[3]

    # @cached_property
    # def labels(self) -> dict:
    #     return pass


def deploy_parameter(
    ssm_client,
    name: str,
    data: T.Union[str, list, dict, T.Any],
    description: T.Optional[str] = None,
    type_is_string: bool = False,
    type_is_string_list: bool = False,
    type_is_secure_string: bool = False,
    kms_key_id: T.Optional[str] = None,
    use_default_kms_key: T.Optional[bool] = False,
    tier_is_standard: bool = False,
    tier_is_advanced: bool = False,
    tier_is_intelligent: bool = False,
    policies: T.Optional[str] = None,
    tags: T.Dict[str, str] = None,
    overwrite: bool = False,
    skip_if_duplicated: bool = True,
) -> T.Optional[Parameter]:
    """
    Create or Update a parameter.

    Ref:

    - put_parameter: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ssm.html#SSM.Client.put_parameter

    Note:

        - you cannot change tags when overwriting parameter, you have to call
            ``add_tags_to_resource`` API after overwriting.

    :param ssm_client: boto3 system manager client
    :param name: the parameter name
    :param data: the parameter data, could be one of the following:
        - a string
        - a list of string
        - a list of object
        - a dict object
        - arbitrary jsonpicklible python object
    :param description: description of the parameter
    :param type_is_string: is it String type?
    :param type_is_string_list: is it StringList type?
    :param type_is_secure_string: is it SecureString type?
    :param kms_key_id: user defined KMS key id for encryption
    :param use_default_kms_key: if true, then you can omit the ``kms_key_id``
        field, and it will use the default ``alias/aws/ssm`` kms key to encrypt
        your data, and the type become Se
    :param tier_is_standard: is this standard tier?
    :param tier_is_advanced: is this advanced tier?
    :param tier_is_intelligent: is this intelligent tier?
    :param policies: access policy
    :param tags: aws resource tags in python dict
    :param overwrite: if False, then raise error when overwriting an existing parameter
    :param skip_if_duplicated: if True, then won't do deployment if parameter data
        is the same as the one in the latest version.

    :return: None or an :class:`Parameter` object, None means that the deployment
        doesn't happen.
    """
    # --------------------------------------------------------------------------
    # input argument pre processing
    # --------------------------------------------------------------------------
    put_parameter_kwargs = {
        "Name": name,
        "Overwrite": overwrite,
    }

    # type
    ensure_only_one_true(
        [
            ("type_is_string", type_is_string),
            ("type_is_string_list", type_is_string_list),
            ("type_is_secure_string", type_is_secure_string),
        ]
    )

    if type_is_string:
        put_parameter_kwargs["Type"] = ParameterTypeEnum.string.value
    if type_is_string_list:
        put_parameter_kwargs["Type"] = ParameterTypeEnum.string_list.value
    if type_is_secure_string:
        put_parameter_kwargs["Type"] = ParameterTypeEnum.secure_string.value

    # value
    if isinstance(data, str):
        put_parameter_kwargs["Value"] = data
    elif isinstance(data, list):
        if type_is_string_list:
            for item in data:
                if "," in item:
                    raise ValueError(
                        "your parameter data has ',' in it, "
                        "you cannot use StringList type, "
                        "please use String type!"
                    )
            put_parameter_kwargs["Value"] = ",".join(data)
        else:
            put_parameter_kwargs["Value"] = json.dumps(data)
    elif isinstance(data, dict):
        put_parameter_kwargs["Value"] = json.dumps(data)
    else:
        if has_jsonpickle is False:  # pragma: no cover
            raise ImportError(
                "you have to install `jsonpickle` to store arbitrary "
                "python object in AWS Parameter Store."
            )
        put_parameter_kwargs["Value"] = json.dumps(
            {JSON_PICKLE_KEY: jsonpickle.dumps(data)}
        )

    # description
    if description is not None:
        put_parameter_kwargs["Description"] = description

    # encryption related
    with_encryption = False
    if (kms_key_id is not None) and (use_default_kms_key is True):
        raise ValueError(
            "you cannot set `kms_key_id` and `use_default_kms_key = True` "
            "at same time!"
        )
    elif (kms_key_id is not None) and (use_default_kms_key is not True):
        if type_is_secure_string is False:
            raise ValueError(
                f"you cannot set kms_key_id = {kms_key_id!r}, "
                f"use_default_kms_key = {use_default_kms_key!r} "
                f"when type is SecureString!"
            )
        put_parameter_kwargs["KeyId"] = kms_key_id  # pragma: no cover
        with_encryption = True  # pragma: no cover
    elif (kms_key_id is None) and (use_default_kms_key is True):
        if type_is_secure_string is False:
            raise ValueError(
                f"you cannot set kms_key_id = {kms_key_id!r}, "
                f"use_default_kms_key = {use_default_kms_key!r} "
                f"when type is SecureString!"
            )
        put_parameter_kwargs["KeyId"] = DEFAULT_KMS_KEY
        with_encryption = True
    elif (kms_key_id is None) and (use_default_kms_key is not True):
        if type_is_secure_string is True:
            put_parameter_kwargs["KeyId"] = DEFAULT_KMS_KEY
            with_encryption = True
    else:  # pragma: no cover
        raise NotImplementedError

    # tier
    ensure_only_one_true(
        [
            ("tier_is_standard", tier_is_standard),
            ("tier_is_advanced", tier_is_advanced),
            ("tier_is_intelligent", tier_is_intelligent),
        ]
    )

    if tier_is_standard:
        put_parameter_kwargs["Tier"] = ParameterTierEnum.standard.value
    if tier_is_advanced:
        put_parameter_kwargs["Tier"] = ParameterTierEnum.advanced.value
    if tier_is_intelligent:
        put_parameter_kwargs["Tier"] = ParameterTierEnum.intelligent_tiering.value

    # policy
    if policies is not None:  # pragma: no cover
        put_parameter_kwargs["Policies"] = policies

    # tag
    if tags is None:
        tags = dict()
    tags_ = [{"Key": k, "Value": v} for k, v in tags.items()]

    # overwrite
    if overwrite:
        put_parameter_kwargs["Overwrite"] = overwrite

    # --------------------------------------------------------------------------
    # create or update
    # --------------------------------------------------------------------------
    # check duplication
    if skip_if_duplicated:
        # check if parameter exists
        parameter = Parameter.load(
            ssm_client=ssm_client,
            name=name,
            with_decryption=with_encryption,
        )
        # if not exists, do create
        if parameter is None:
            if len(tags_):
                put_parameter_kwargs["Tags"] = tags_
            response = ssm_client.put_parameter(**put_parameter_kwargs)
            return Parameter._from_put_parameter_response(
                put_parameter_kwargs, response
            )
        # if already exists, compare the parameter data
        else:
            # if the same, do nothing
            if parameter.Value == put_parameter_kwargs["Value"]:
                return None
            # if not same, do update
            else:
                response = ssm_client.put_parameter(**put_parameter_kwargs)
                if len(tags_):
                    ssm_client.add_tags_to_resource(
                        ResourceType="Parameter",
                        ResourceId=name,
                        Tags=tags_,
                    )
                return Parameter._from_put_parameter_response(
                    put_parameter_kwargs, response
                )
    # don't duplication check, just update
    else:
        response = ssm_client.put_parameter(**put_parameter_kwargs)
        if len(tags_):
            ssm_client.add_tags_to_resource(
                ResourceType="Parameter",
                ResourceId=name,
                Tags=tags_,
            )
        return Parameter._from_put_parameter_response(put_parameter_kwargs, response)


def delete_parameter(
    ssm_client,
    name: str,
) -> bool:
    """
    Delete a Parameter.

    Ref:

    - delete_parameter: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/ssm.html#SSM.Client.delete_parameter

    :return: a boolean value to indicate whether a deletion happened.
    """
    try:
        ssm_client.delete_parameter(Name=name)
        return True
    except Exception as e:
        if "ParameterNotFound" in str(e):
            return False
        else:  # pragma: no cover
            raise e



# -*- coding: utf-8 -*-

import weakref


class CachedSpam(object):
    """
    Cache instance with based on its' name.
    """
    settings_uuid_field = None

    _cache = None

    def __init__(self, *args, **kwargs):
        msg = "Can't instantiate directly, use {}.new(...) instead." \
            .format(self.__class__.__name__)
        raise RuntimeError(msg)

    def __real_init__(self, *args, **kwargs):
        msg = ("You have to implement ``def __real_init__(self, ...)`` "
               "like you implement ``def __init__(self, ...)``!")
        raise NotImplementedError(msg)

    @classmethod
    def _init_cache(cls):
        if cls._cache is None:
            cls._cache = weakref.WeakValueDictionary()

    @classmethod
    def new(cls, *args, **kwargs):
        cls._init_cache()

        if cls.settings_uuid_field is None:
            msg = "You have to specify `CachedSpam.settings_uuid_field`!"
            raise NotImplementedError(msg)

        if cls.settings_uuid_field not in kwargs:
            msg = "Can't find '{}' in {} Please use `CachedSpam.new({}=xxx)`".format(
                cls.settings_uuid_field, kwargs, cls.settings_uuid_field)
            raise SyntaxError(msg)

        uuid = kwargs[cls.settings_uuid_field]
        if uuid in cls._cache:
            return cls._cache[uuid]
        else:
            self = cls.__new__(cls)
            self.__real_init__(*args, **kwargs)
            cls._cache[uuid] = self
            return self



# -*- coding: utf-8 -*-

import subprocess

from .paths import bin_pytest, dir_htmlcov, dir_project_root, path_cov_index_html


def _run_cov_test(
    bin_pytest: str,
    script: str,
    module: str,
    root_dir: str,
    htmlcov_dir: str,
):
    """
    A simple wrapper around pytest + coverage cli command.

    :param bin_pytest: the path to pytest executable
    :param script: the path to test script
    :param module: the dot notation to the python module you want to calculate
        coverage
    :param root_dir: the dir to dump coverage results binary file
    :param htmlcov_dir: the dir to dump HTML output
    """
    args = [
        bin_pytest,
        "-s",
        "--tb=native",
        f"--rootdir={root_dir}",
        f"--cov={module}",
        "--cov-report",
        "term-missing",
        "--cov-report",
        f"html:{htmlcov_dir}",
        script,
    ]
    subprocess.run(args)


def run_cov_test(script: str, module: str, preview: bool = False):
    _run_cov_test(
        bin_pytest=f"{bin_pytest}",
        script=script,
        module=module,
        root_dir=f"{dir_project_root}",
        htmlcov_dir=f"{dir_htmlcov}",
    )
    if preview:
        subprocess.run(["open", f"{path_cov_index_html}"])



# -*- coding: utf-8 -*-

"""
AWS Key Management Service supportintegration
"""


def kms_symmetric_encrypt(
    kms_client,
    blob: bytes,
    kms_key_id: str,
):
    """
    Use KMS key to encrypt a short text.

    - KMS.Client.encrypt: https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/kms.html#KMS.Client.encrypt

    :param blob: binary data to encrypt
    :param kms_key_id:

    :rtype: bytes
    """
    return kms_client.encrypt(
        Plaintext=blob,
        KeyId=kms_key_id,
    )["CiphertextBlob"]


def kms_symmetric_decrypt(
    kms_client,
    blob: bytes,
):
    """
    Use KMS key to decrypt a short text.

    :param blob: binary data to decrypt

    :rtype: bytes
    """
    return kms_client.decrypt(CiphertextBlob=blob)["Plaintext"]



# -*- coding: utf-8 -*-

import typing as T
import os
import json
import dataclasses
from pathlib import Path

from .sh_helper import load_var_value_from_shell_script


@dataclasses.dataclass
class BaseShellScriptSecret:
    """
    A base class for environment secrets data container.
    """

    @classmethod
    def load(cls, path_sh: str):
        data = load_var_value_from_shell_script(path_sh)
        kwargs = {f.name: data.get(f.name) for f in dataclasses.fields(cls)}
        return cls(**kwargs)

    def to_json(self, pretty: bool = True) -> str:
        kwargs = dict()
        if pretty:
            kwargs["indent"] = 4
        return json.dumps(dataclasses.asdict(self), **kwargs)
